"""AD9680 high speed ADC clocking model."""

from typing import Any, Dict, List, Union

from adijif.converters.ad9680_bf import ad9680_bf

from ..solvers import CpoSolveResult  # noqa: I202
from .ad9680_draw import ad9680_draw  # noqa: I202


def _convert_to_config(
    L: Union[int, float],
    M: Union[int, float],
    F: Union[int, float],
    S: Union[int, float],
    HD: Union[int, float],
    N: Union[int, float],
    Np: Union[int, float],
    CS: Union[int, float],
) -> Dict:
    # return {"L": L, "M": M, "F": F, "S": S, "HD": HD, "N": N, "Np": Np, "CS": CS}
    return {
        "L": L,
        "M": M,
        "F": F,
        "S": S,
        "HD": HD,
        "Np": Np,
        "jesd_class": "jesd204b",
    }


quick_configuration_modes = {
    # M = 1
    str(0x01): _convert_to_config(1, 1, 2, 1, 0, -1, 16, -1),
    str(0x40): _convert_to_config(2, 1, 1, 1, 1, -1, 16, -1),
    str(0x41): _convert_to_config(2, 1, 2, 2, 0, -1, 16, -1),
    str(0x80): _convert_to_config(4, 1, 1, 2, 1, -1, 16, -1),
    str(0x81): _convert_to_config(4, 1, 2, 4, 0, -1, 16, -1),
    # M = 2
    str(0x0A): _convert_to_config(1, 2, 4, 1, 0, -1, 16, -1),
    str(0x49): _convert_to_config(2, 2, 2, 1, 0, -1, 16, -1),
    str(0x88): _convert_to_config(4, 2, 1, 1, 1, -1, 16, -1),
    str(0x89): _convert_to_config(4, 2, 2, 2, 0, -1, 16, -1),
    # M = 4
    str(0x13): _convert_to_config(1, 4, 8, 1, 0, -1, 16, -1),
    str(0x52): _convert_to_config(2, 4, 4, 1, 0, -1, 16, -1),
    str(0x91): _convert_to_config(4, 4, 2, 1, 0, -1, 16, -1),
    # M = 8
    str(0x1C): _convert_to_config(1, 8, 16, 1, 0, -1, 16, -1),
    str(0x5B): _convert_to_config(2, 8, 8, 1, 0, -1, 16, -1),
    str(0x9A): _convert_to_config(4, 8, 4, 1, 0, -1, 16, -1),
}


class ad9680(ad9680_draw, ad9680_bf):
    """AD9680 high speed ADC model.

    This model supports direct clock configurations

    Clocking: AD9680 has directly clocked ADC that have optional input dividers.
    The sample rate can be determined as follows:

        baseband_sample_rate = (input_clock / input_clock_divider) / datapath_decimation
    """

    name = "AD9680"
    converter_type = "ADC"

    # JESD parameters
    _jesd_params_to_skip_check = ["DualLink", "CS", "N", "HD"]
    available_jesd_modes = ["jesd204b"]
    K_available = [4, 8, 12, 16, 20, 24, 28, 32]
    L_available = [1, 2, 4]
    M_available = [1, 2, 4, 8]
    N_available = [*range(7, 16)]
    Np_available = [8, 16]
    F_available = [1, 2, 4, 8, 16]
    CS_available = [0, 1, 2, 3]
    CF_available = [0]
    S_available = [1, 2, 4]

    # Clock constraints
    clocking_option_available = ["direct"]
    _clocking_option = "direct"
    converter_clock_min = 300e6
    converter_clock_max = 1.25e9
    bit_clock_min_available = {"jesd204b": 3.125e9}
    bit_clock_max_available = {"jesd204b": 12.5e9}
    sample_clock_min = 300e6
    sample_clock_max = 1250e6
    decimation_available = [1, 2, 4, 8, 16]
    _decimation = 1
    _lmfc_sys_divisor = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    quick_configuration_modes = {"jesd204b": quick_configuration_modes}

    # Input clock requirements
    input_clock_divider_available = [1, 2, 4, 8]  # FIXME
    input_clock_divider = 1  # FIXME
    input_clock_max = 4e9  # FIXME

    """ Clocking
        AD9680 has directly clocked ADCs that have optional input dividers.
        The sample rate can be determined as follows:

        baseband_sample_rate = (input_clock / input_clock_divider) / datapath_decimation
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize AD9680 class.

        Objects will default to mode 0x88 with 1e9 sample_clock.

        Args:
            *args (Any): Pass through arguments
            **kwargs (Any): Pass through keyword arguments
        """
        # Set default mode
        self.set_quick_configuration_mode(str(0x88))
        self.K = 32
        self.sample_clock = 1e9
        super().__init__(*args, **kwargs)
        self._init_diagram()

    def _check_valid_jesd_mode(self) -> str:
        """Verify current JESD configuration for part is valid.

        Returns:
            str: Current JESD mode
        """
        if self.F == 1:
            assert self.K in [20, 24, 28, 32], "Invalid K value for F=1"
        if self.F == 2:
            assert self.K in [12, 16, 20, 24, 28, 32], "Invalid K value for F=1"
        if self.F == 4:
            assert self.K in [8, 12, 16, 20, 24, 28, 32], "Invalid K value for F=1"
        if self.F in [8, 16]:
            assert self.K in [4, 8, 12, 16, 20, 24, 28, 32], "Invalid K value for F=1"

        return super()._check_valid_jesd_mode()

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal converter configuration and output clock definitions
        leading to connected devices (clock chips, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration
        """
        if solution:
            self._saved_solution = solution
        return {"clocking_option": self.clocking_option, "decimation": self.decimation}

    def get_required_clock_names(self) -> List[str]:
        """Get list of strings of names of requested clocks.

        This list of names is for the clocks defined by get_required_clocks

        Returns:
            List[str]: List of strings of clock names in order
        """
        return ["AD9680_ref_clk", "AD9680_sysref"]

    def get_required_clocks(self) -> List:
        """Generate list required clocks.

        For AD9680 this will contain [converter clock, sysref requirement SOS]

        Returns:
            List: List of solver variables, equations, and constants
        """
        # possible_sysrefs = []
        # for n in range(1, 10):
        #     r = self.multiframe_clock / (n * n)
        #     if r == int(r) and r > 1e6:
        #         possible_sysrefs.append(r)
        # self.config = {"sysref": self.model.sos1(possible_sysrefs)}

        self.config = {}
        self.config["lmfc_divisor_sysref"] = self._convert_input(
            self._lmfc_sys_divisor,
            default=self._lmfc_sys_divisor[-1],
            name="AD9680_lmfc_divisor_sysref",
        )

        self.config["lmfc_divisor_sysref_squared"] = self._add_intermediate(
            self.config["lmfc_divisor_sysref"] * self.config["lmfc_divisor_sysref"]
        )
        self.config["sysref"] = self._add_intermediate(
            self.multiframe_clock / self.config["lmfc_divisor_sysref_squared"]
        )

        # Objectives
        # self.model.Obj(self.config["sysref"])  # This breaks many searches

        return [self.converter_clock, self.config["sysref"]]



"""AD9545 clock chip model."""

from typing import Dict, List

import docplex.cp.catalog as ctg
import docplex.cp.expression as exp
import docplex.cp.modeler as mod

from adijif.clocks.clock import clock
from adijif.solvers import CpoSolveResult


class ad9545(clock):
    """AD9545 clock chip model.

    Currently this model supports only the internal zero delay mode

    PLL_in_rate_0 = input_ref_0 / r0
    PLL_in_rate_1 = input_ref_1 / r1
    PLL_in_rate_2 = input_ref_2 / r2
    PLL_in_rate_3 = input_ref_3 / r3

    PLL_in_rate_0 < PLL_in_max
    PLL_in_rate_1 < PLL_in_max
    PLL_in_rate_2 < PLL_in_max
    PLL_in_rate_3 < PLL_in_max

    PLL0_rate = n0_profile_0 * PLL_in_rate_0
    PLL0_rate = n0_profile_1 * PLL_in_rate_1
    PLL0_rate = n0_profile_2 * PLL_in_rate_2
    PLL0_rate = n0_profile_3 * PLL_in_rate_3

    out_rate_0 = PLL0_rate / q0
    out_rate_1 = PLL0_rate / q1
    ...
    out_rate_5 = PLL0_rate / q5
    """

    # Limits
    """ Internal limits """
    PLL_out_min = [1200e6, 1600e6]
    PLL_out_max = [1600e6, 2000e6]

    PLL_in_min = 1
    PLL_in_max = 200000

    APLL_PFD_MIN = int(162e6)
    APLL_PFD_MAX = int(350e6)

    """ Output dividers limits

        Value is programmed on 32 bits but
        2 GHz is the maximum PLL output that feeds
        directly to the output dividers.
    """
    Q_max = 2000000000
    Q_min = 1

    """ Input dividers limits

        Value is programmed on 32 bits but
        750 MHz is the maximum input frequency anyway.
    """
    R_max = 750000000
    R_min = 1

    """ PLL N dividers limits

        Value is programmed on 32 bits.

        There are constraints on the input and output frequency.
        This places lower and upper bounds on N too.

        Maximum input frequency is 200 kHz.
    """
    N_max = 2000000000
    N_min = 1

    PLL_used = [False, False]

    avoid_min_max_PLL_rates = True
    minimize_input_dividers = True

    """ Hitless mode

        Output phase is alligned with the input phase.

        If true, DPLL feedback path will no longer be
        the one from APLL output. It will be from a designated
        output divider.

        These outputs that are used as feedback for DPLL are
        assigned to a specific PLL profile.

        f_out_fb_source * r_div = dpll_i_N * f_input_ref
    """
    profiles = {
        "dpll_0_profile_0": {"hitless": False, "fb_source": 0},
        "dpll_0_profile_1": {"hitless": False, "fb_source": 0},
        "dpll_0_profile_2": {"hitless": False, "fb_source": 0},
        "dpll_0_profile_3": {"hitless": False, "fb_source": 0},
        "dpll_1_profile_0": {"hitless": False, "fb_source": 0},
        "dpll_1_profile_1": {"hitless": False, "fb_source": 0},
        "dpll_1_profile_2": {"hitless": False, "fb_source": 0},
        "dpll_1_profile_3": {"hitless": False, "fb_source": 0},
    }

    config = {
        "r0": 0,
        "r1": 0,
        "r2": 0,
        "r3": 0,
        "PLL0": {},
        "PLL1": {},
        "q0": 0,
        "q1": 0,
        "q2": 0,
        "q3": 0,
        "q4": 0,
        "q5": 0,
        "q6": 0,
        "q7": 0,
        "q8": 0,
        "q9": 0,
    }

    def list_available_references(self) -> List[int]:
        """List available references for a given divider set."""
        return [self.ref0, self.ref1, self.ref2, self.ref3]

    def find_dividers(self) -> None:
        """Find dividers for a given input reference.

        Not implemented for this model.

        Raises:
            NotImplementedError: Always
        """
        raise NotImplementedError

    def get_config(self, solution: CpoSolveResult = None) -> Dict:
        """Extract configurations from solver results.

        Collect internal clock chip configuration and output clock definitions
        leading to connected devices (converters, FPGAs)

        Args:
            solution (CpoSolveResult): CPlex solution. Only needed for CPlex solver

        Returns:
            Dict: Dictionary of clocking rates and dividers for configuration
        """
        if solution:
            self.solution = solution

        config: Dict = {
            "PLL0": {},
            "PLL1": {},
        }

        for i in range(0, 4):
            config["r" + str(i)] = self._get_val(self.config["r" + str(i)])

        for i in range(0, 2):
            for j in range(0, 4):
                if self.input_refs[j] != 0 and self.PLL_used[i]:
                    n_name = "n" + str(i) + "_profile_" + str(j)
                    n_dpll_name = "n_dpll" + str(i) + "_profile_" + str(j)

                    config["PLL" + str(i)][n_dpll_name] = self._get_val(
                        self.config[n_dpll_name]
                    )

                    config["PLL" + str(i)][n_name] = self._get_val(self.config[n_name])

                    config["PLL" + str(i)]["rate_hz"] = self._get_val(
                        self.config["PLL" + str(i) + "_rate"]
                    )

        for i in range(0, 2):
            for j in range(0, 4):
                dpll_profile_name = "dpll_" + str(i) + "_profile_" + str(j)

                if self.PLL_used[i] and dpll_profile_name in self.profiles:
                    if self.profiles[dpll_profile_name]["hitless"]:
                        source_nr = int(self.profiles[dpll_profile_name]["fb_source"])

                        config["PLL" + str(i)]["hitless"] = {
                            "fb_source": source_nr,
                            "fb_source_rate": int(self.out_freqs[source_nr]),
                        }
                        break

        for i in range(0, 10):
            config["q" + str(i)] = self._get_val(self.config["q" + str(i)])

        return config

    def _setup_solver_constraints(
        self, input_refs: List[int], out_freqs: List[int]
    ) -> None:
        """Apply constraints to solver model.

        Args:
            input_refs (List[int]): 4 references (frequency in hertz)
            out_freqs (List[int]): 10 outputs (frequency in hertz)

        Raises:
            Exception: Invalid solver
        """
        self.input_refs = input_refs

        self.PLL_used = [False, False]
        for i in range(0, 10):
            if out_freqs[i] != 0:
                if i > 5:
                    self.PLL_used[1] = True
                else:
                    self.PLL_used[0] = True

        equations = []

        if self.solver == "gekko":
            """Add divider as variables to the model"""
            for i in range(0, 4):
                if input_refs[i] != 0:
                    """If the user does not set a specific r,
                    turn it into a variable"""
                    r_div_int = isinstance(self.config["r" + str(i)], int)
                    if r_div_int and self.config["r" + str(i)] != 0:
                        self.config["r" + str(i)] = self.model.Const(
                            int(self.config["r" + str(i)]),
                            name=("r" + str(i)),
                        )
                    else:
                        self.config["r" + str(i)] = self.model.Var(
                            integer=True,
                            lb=self.R_min,
                            ub=self.R_max,
                            name=("r" + str(i)),
                        )
                    self.config["input_ref_" + str(i)] = self.model.Const(
                        int(input_refs[i]),
                        name=("input_ref_" + str(i)),
                    )

            """ Add PLL N as variables to the model for each PLL profile """
            for i in range(0, 2):
                if self.PLL_used[i]:
                    """Force PLL input rates and output rates
                    to be integer values"""
                    self.config["PLL" + str(i) + "_rate"] = self.model.Var(
                        integer=True,
                        lb=self.PLL_out_min[i],
                        ub=self.PLL_out_max[i],
                        name=("PLL" + str(i) + "_rate"),
                    )

                    for j in range(0, 4):
                        if input_refs[j] != 0:
                            n_name = "n" + str(i) + "_profile_" + str(j)
                            n_dpll_name = "n_dpll" + str(i) + "_profile_" + str(j)
                            m_apll_name = "m_apll" + str(i) + "_profile_" + str(j)

                            self.config[n_name] = self.model.Var(
                                integer=True,
                                lb=self.N_min,
                                ub=self.N_max,
                                name=n_name,
                            )

                            """ Internally the PLL block is composed of a
                            Digital PLL and an Analog PLL with different
                            constraints on dividers.
                            """
                            DPLL_N = self.model.Var(
                                integer=True, lb=1, ub=350e6, name=n_dpll_name
                            )
                            self.config[n_dpll_name] = DPLL_N

                            APLL_M = self.model.Var(
                                integer=True, lb=7, ub=255, name=m_apll_name
                            )
                            self.config[m_apll_name] = APLL_M

                            equations = equations + [
                                self.config[n_name] == DPLL_N * APLL_M
                            ]

        elif self.solver == "CPLEX":
            """Add divider as variables to the model"""
            for i in range(0, 4):
                if input_refs[i] != 0:
                    """If the user does not set a specific r,
                    turn it into a variable
                    """

                    r_div_int = isinstance(self.config["r" + str(i)], int)
                    if r_div_int and self.config["r" + str(i)] != 0:
                        self.config["input_ref_" + str(i)] = exp.CpoValue(
                            int(self.config["r" + str(i)]), type=ctg.Type_Int
                        )
                    else:
                        self.config["r" + str(i)] = exp.integer_var(
                            int(self.R_min), int(self.R_max), "r" + str(i)
                        )
                    self.config["input_ref_" + str(i)] = exp.CpoValue(
                        int(input_refs[i]), type=ctg.Type_Int
                    )

            """ Add PLL N as variables to the model for each PLL profile """
            for i in range(0, 2):
                if self.PLL_used[i]:
                    self.config["PLL" + str(i) + "_rate"] = exp.integer_var(
                        int(self.PLL_out_min[i]),
                        int(self.PLL_out_max[i]),
                        "PLL" + str(i) + "_rate",
                    )

                    for j in range(0, 4):
                        if input_refs[j] != 0:
                            n_name = "n" + str(i) + "_profile_" + str(j)
                            n_dpll_name = "n_dpll" + str(i) + "_profile_" + str(j)
                            m_apll_name = "m_apll" + str(i) + "_profile_" + str(j)

                            self.config[n_name] = exp.integer_var(
                                int(self.N_min), int(self.N_max), n_name
                            )

                            """ Internally the PLL block is composed of a Digital
                            PLL and an Analog PLL with different constraints on
                            dividers
                            """
                            DPLL_N = exp.integer_var(int(1), int(350e6), n_dpll_name)
                            self.config[n_dpll_name] = DPLL_N
                            APLL_M = exp.integer_var(int(7), int(255), m_apll_name)
                            self.config[m_apll_name] = APLL_M

                            equations = equations + [
                                self.config[n_name] == DPLL_N * APLL_M
                            ]
        else:
            raise Exception("Unknown solver {}".format(self.solver))

        for i in range(0, 4):
            if self.input_refs[i] != 0:
                if self.solver == "gekko":
                    self.config["PLL_in_rate_" + str(i)] = self.model.Var(
                        integer=True,
                        lb=self.PLL_in_min,
                        ub=self.PLL_in_max,
                        name=("PLL_in_rate_" + str(i)),
                    )
                elif self.solver == "CPLEX":
                    self.config["PLL_in_rate_" + str(i)] = exp.integer_var(
                        int(self.PLL_in_min),
                        int(self.PLL_in_max),
                        "PLL_in_rate_" + str(i),
                    )
                else:
                    raise Exception("Unknown solver {}".format(self.solver))

                equations = equations + [
                    self.config["PLL_in_rate_" + str(i)] * self.config["r" + str(i)]
                    == self.config["input_ref_" + str(i)]
                ]

        for i in range(0, 2):
            for j in range(0, 4):
                if self.input_refs[j] != 0 and self.PLL_used[i]:
                    n_name = "n" + str(i) + "_profile_" + str(j)
                    n_dpll_name = "n_dpll" + str(i) + "_profile_" + str(j)
                    m_apll_name = "m_apll" + str(i) + "_profile_" + str(j)
                    dpll_profile_name = "dpll_" + str(i) + "_profile_" + str(j)

                    pll_rate = self.config["PLL" + str(i) + "_rate"]
                    pll_in_rate = self.config["PLL_in_rate_" + str(j)]
                    pll_m_div = self.config[m_apll_name]
                    pll_n_div = self.config[n_dpll_name]

                    if not self.profiles[dpll_profile_name]["hitless"]:
                        equations = equations + [
                            self.config[n_name] * pll_in_rate == pll_rate
                        ]

                        """ Limit APLL PFD input values """
                        equations = equations + [
                            (pll_n_div * pll_in_rate) >= self.APLL_PFD_MIN,
                            (pll_n_div * pll_in_rate) <= self.APLL_PFD_MAX,
                        ]
                    else:
                        """Limit APLL PFD input values"""
                        equations = equations + [
                            pll_rate >= self.APLL_PFD_MIN * pll_m_div,
                            pll_rate <= self.APLL_PFD_MAX * pll_m_div,
                        ]

        self._add_equation(equations)

        cplex_objectives = []

        """ Instruct solver to try to avoid Minimum and Maximum PLL rates """
        if self.avoid_min_max_PLL_rates:
            for i in range(0, 2):
                if self.PLL_used[i]:
                    average_PLL_rate = self.PLL_out_min[i] / 2 + self.PLL_out_max[i] / 2

                    if self.solver == "CPLEX":
                        cplex_objectives = cplex_objectives + [
                            mod.abs(
                                self.config["PLL" + str(i) + "_rate"] - average_PLL_rate
                            )
                        ]
                    elif self.solver == "gekko":
                        self.model.Minimize(
                            self.model.abs3(
                                self.config["PLL" + str(i) + "_rate"] - average_PLL_rate
                            )
                        )
                    else:
                        raise Exception("Unknown solver {}".format(self.solver))

        """ Instruct solver to maximize PLL input rates """
        if self.minimize_input_dividers:
            for i in range(0, 4):
                if self.input_refs[i] != 0:
                    if self.solver == "CPLEX":
                        cplex_objectives = cplex_objectives + [
                            self.config["r" + str(i)]
                        ]
                    elif self.solver == "gekko":
                        self.model.Maximize(self.config["PLL_in_rate_" + str(i)])
                    else:
                        raise Exception("Unknown solver {}".format(self.solver))

        if self.solver == "CPLEX" and len(cplex_objectives) != 0:
            self.model.add(mod.minimize_static_lex(cplex_objectives))

    def _setup(self, input_refs: List[int], out_freqs: List[int]) -> None:
        # Setup clock chip internal constraints
        self.out_freqs = out_freqs

        self._setup_solver_constraints(input_refs, out_freqs)

    def set_requested_clocks(self, ins: List[int], outs: List[int]) -> None:
        """Define necessary clocks to be generated in model.

        Args:
            ins (List[int]): list of input references rates
            outs (List[int]): list of output rates required

        Raises:
            Exception: if the number of input references is not equal to the
                number of output rates
        """
        input_refs = [0] * 4
        out_freqs = [0] * 10

        for in_ref in ins:
            input_refs[in_ref[0]] = in_ref[1]

        for out_ref in outs:
            out_freqs[out_ref[0]] = out_ref[1]

        self._setup(input_refs, out_freqs)

        """ Add output dividers as variables to the model """
        for i in range(0, 10):
            if out_freqs[i] != 0:
                if self.solver == "gekko":
                    self.config["q" + str(i)] = self.model.Var(
                        integer=True,
                        lb=self.Q_min,
                        ub=self.Q_max,
                        name=("q" + str(i)),
                    )
                    self.config["out_rate_" + str(i)] = self.model.Const(
                        int(out_freqs[i]), name=("out_rate_" + str(i))
                    )
                elif self.solver == "CPLEX":
                    self.config["q" + str(i)] = exp.integer_var(
                        int(self.Q_min), int(self.Q_max), "q" + str(i)
                    )
                    self.config["out_rate_" + str(i)] = exp.CpoValue(
                        value=int(out_freqs[i]), type=ctg.Type_Int
                    )
                else:
                    raise Exception("Unknown solver {}".format(self.solver))

        """ Add hitless mode constraints for profiles that activate it """
        for i in range(0, 2):
            for j in range(0, 4):
                dpll_profile_name = "dpll_" + str(i) + "_profile_" + str(j)

                if self.PLL_used[i] and self.profiles[dpll_profile_name]["hitless"]:
                    source_nr = int(self.profiles[dpll_profile_name]["fb_source"])
                    n_dpll_name = "n_dpll" + str(i) + "_profile_" + str(j)

                    if out_freqs[source_nr] == 0:
                        raise Exception(
                            "Hitless profile: {}, has an invalid source.".format(
                                dpll_profile_name
                            )
                        )

                    input_ref = self.config["input_ref_" + str(j)]
                    pll_n_div = self.config[n_dpll_name]
                    out_rate = self.config["out_rate_" + str(source_nr)]
                    r_div = self.config["r" + str(j)]

                    """ Frequency translation factor:
                    N * input_ref_j == out_rate_x * r_div_j
                    """
                    self._add_equation([input_ref * pll_n_div == out_rate * r_div])

                    """ Hitless mode places a strict constraint on Q dividers """
                    self.config["q" + str(i)]
                    self._add_equation([self.config["q" + str(source_nr)] >= 8])

        for i in range(0, 10):
            if out_freqs[i] != 0:
                if i > 5:
                    pll_number = 1
                else:
                    pll_number = 0

                self._add_equation(
                    [
                        self.config["PLL" + str(pll_number) + "_rate"]
                        == self.config["out_rate_" + str(i)] * self.config["q" + str(i)]
                    ]
                )

    def _solve_gekko(self) -> bool:
        """Local solve method for clock model.

        Call model solver with correct arguments.

        Returns:
            bool: Always False
        """
        self.model.options.SOLVER = 1  # APOPT solver
        self.model.solver_options = [
            "minlp_maximum_iterations 300000000",
            "minlp_max_iter_with_int_sol 6000",
            "minlp_as_nlp 0",
            "minlp_branch_method 1",
            "minlp_integer_tol 0.0001",
        ]

        self.model.solve(disp=False)

        return False



"""Main entry point for PyADI-JIF Tools Explorer Streamlit application."""

import logging
import os
from typing import Optional

import streamlit as st
from src.pages import PAGE_MAP
from src.utils import add_custom_css

# from src.state import provide_state


# Global logging configuration
logging.basicConfig(level=logging.ERROR)

# Configure page settings - must be first Streamlit command
st.set_page_config(
    page_title="PyADI-JIF Tools Explorer",
    page_icon=os.path.join(os.path.dirname(__file__), "PyADI-JIF_logo.png"),
    initial_sidebar_state="expanded",
)

# Apply custom CSS to main page and sidebar
add_custom_css()


# @provide_state()
def main(state: Optional[object] = None) -> None:
    """Run the main Streamlit application."""
    logo_image = os.path.join(os.path.dirname(__file__), "PyADI-JIF_logo.png")
    st.sidebar.image(logo_image)
    st.sidebar.title("Tools Explorer")
    current_page = st.sidebar.radio(
        "Select a Tool", list(PAGE_MAP), label_visibility="hidden"
    )
    PAGE_MAP[current_page](state=state).write()


if __name__ == "__main__":
    main()



# flake8: noqa
import pytest

import adijif as jif


@pytest.mark.parametrize(
    "conv",
    [
        "ad9680",
        "adrv9009_rx",
        "adrv9009_tx",
        "ad9081_rx",
        "ad9081_tx",
        "ad9144",
        "ad9084_rx",
        "ad9088_rx",
        "ad9082_rx",
        "ad9082_tx",
    ],
)
def test_converters(conv):
    c = eval(f"jif.{conv}()")
    # Check static
    c.validate_config()

    required_clocks = c.get_required_clocks()
    required_clock_names = c.get_required_clock_names()

    # Add generic clock sources for solver
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):
        clk = jif.types.arb_source(name)
        c._add_equation(clk(c.model) == clock)
        clks.append(clk)

    # Solve
    solution = c.model.solve(LogVerbosity="Quiet")
    settings = c.get_config(solution)

    # Get clock values
    clock_values = {}
    for clk in clks:
        clock_values.update(clk.get_config(solution))
    settings["clocks"] = clock_values

    print(settings)

    image_data = c.draw(settings["clocks"])
    assert image_data is not None


@pytest.mark.drawing
@pytest.mark.parametrize("clock", ["hmc7044", "ltc6952", "ad9528", "ad9523_1"])
def test_clock_chip_draw(clock):

    import pprint

    vcxo = 125000000

    clk = eval(f"jif.{clock}()")
    # clk = jif.hmc7044()

    output_clocks = [1e9, 500e6, 7.8125e6]
    output_clocks = list(map(int, output_clocks))  # force to be ints
    clock_names = ["ADC", "FPGA", "SYSREF"]

    clk.set_requested_clocks(vcxo, output_clocks, clock_names)

    clk.solve()

    o = clk.get_config()

    pprint.pprint(o)

    clk.draw()


@pytest.mark.drawing
def test_ad9680_draw():
    adc = jif.ad9680()

    # Check static
    adc.validate_config()

    required_clocks = adc.get_required_clocks()
    required_clock_names = adc.get_required_clock_names()

    # Add generic clock sources for solver
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):
        clk = jif.types.arb_source(name)
        adc._add_equation(clk(adc.model) == clock)
        clks.append(clk)

    # Solve
    solution = adc.model.solve(LogVerbosity="Quiet")
    settings = adc.get_config(solution)

    # Get clock values
    clock_values = {}
    for clk in clks:
        clock_values.update(clk.get_config(solution))
    settings["clocks"] = clock_values

    print(settings)

    image_data = adc.draw(settings["clocks"])
    assert image_data is not None

    # with open("ad9680_example.svg", "w") as f:
    #     f.write(image_data)


@pytest.mark.drawing
def test_generic_converter_draw():
    adc = jif.adrv9009_rx()
    adc.sample_clock = 122.88e6
    adc.decimation = 4
    adc.set_quick_configuration_mode("17", "jesd204b")

    # Check static
    adc.validate_config()

    required_clocks = adc.get_required_clocks()
    required_clock_names = adc.get_required_clock_names()

    # Add generic clock sources for solver
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):
        clk = jif.types.arb_source(name)
        adc._add_equation(clk(adc.model) == clock)
        clks.append(clk)

    # Solve
    solution = adc.model.solve(LogVerbosity="Quiet")
    settings = adc.get_config(solution)

    # Get clock values
    clock_values = {}
    for clk in clks:
        clock_values.update(clk.get_config(solution))
    settings["clocks"] = clock_values

    print(settings)

    image_data = adc.draw(settings["clocks"])
    assert image_data is not None


@pytest.mark.drawing
@pytest.mark.parametrize("device_clock_source", ["external", "link_clock", "ref_clock"])
def test_xilinx_draw(device_clock_source):
    import adijif as jif
    from adijif.converters.converter import converter

    fpga = jif.xilinx()
    fpga.setup_by_dev_kit_name("vcu118")
    fpga.device_clock_source = device_clock_source

    # class dummy_converter(converter):
    #     name = "dummy"

    # dc = dummy_converter()
    dc = jif.ad9680()

    fpga_ref = jif.types.arb_source("FPGA_REF")
    link_out_ref = jif.types.arb_source("LINK_OUT_REF")

    clocks = fpga.get_required_clocks(
        dc, fpga_ref(fpga.model), link_out_ref(fpga.model)
    )
    # print(clocks)

    solution = fpga.model.solve(LogVerbosity="Quiet")
    # solution.write()

    settings = {}
    # Get clock values
    clock_values = {}
    for clk in [fpga_ref, link_out_ref]:
        clock_values.update(clk.get_config(solution))
    settings["clocks"] = clock_values

    settings["fpga"] = fpga.get_config(dc, settings["clocks"]["FPGA_REF"], solution)
    print(settings)

    image_data = fpga.draw(settings)

    del fpga

    assert image_data


@pytest.mark.drawing
def test_system_draw():
    import pprint

    import adijif

    vcxo = 125000000

    sys = adijif.system("ad9680", "hmc7044", "xilinx", vcxo)

    # Get Converter clocking requirements
    sys.converter.sample_clock = 1e9
    sys.converter.decimation = 1
    sys.converter.set_quick_configuration_mode(str(0x88))
    sys.converter.K = 32
    sys.Debug_Solver = False

    # Get FPGA clocking requirements
    sys.fpga.setup_by_dev_kit_name("zc706")
    sys.fpga.force_qpll = 1

    cfg = sys.solve()

    pprint.pprint(cfg)

    # print("Clock config:")
    # pprint.pprint(cfg["clock"])

    # print("Converter config:")
    # pprint.pprint(cfg["converter"])

    print("FPGA config:")
    pprint.pprint(cfg["fpga_AD9680"])

    # print("JESD config:")
    # pprint.pprint(cfg["jesd_AD9680"])

    data = sys.draw(cfg)

    # with open("daq2_example.svg", "w") as f:
    #     f.write(data)


@pytest.mark.drawing
def test_ad9084_draw():
    """Test AD9084 drawing."""
    conv = jif.ad9084_rx()

    # Check static
    conv.validate_config()

    required_clocks = conv.get_required_clocks()
    required_clock_names = conv.get_required_clock_names()

    # Add generic clock sources for solver
    clks = []
    for clock, name in zip(required_clocks, required_clock_names):
        clk = jif.types.arb_source(name)
        conv._add_equation(clk(conv.model) == clock)
        clks.append(clk)

    # Solve
    solution = conv.model.solve(LogVerbosity="Quiet")
    settings = conv.get_config(solution)

    # Get clock values
    clock_values = {}
    for clk in clks:
        clock_values.update(clk.get_config(solution))
    settings["clocks"] = clock_values

    image_data = conv.draw(settings["clocks"])
    assert image_data is not None



"""JESD204 mode selector page."""

from typing import Optional

import pandas as pd
import streamlit as st

from adijif.converters import supported_parts as sp

from ..utils import Page
from .helpers.datapath import gen_datapath
from .helpers.drawers import draw_adc, draw_dac
from .helpers.jesd import get_jesd_controls, get_valid_jesd_modes

# options_to_skip = ["global_index", "decimations"]


class JESDModeSelector(Page):
    """JESD204 mode selector tool page."""

    def __init__(self, state: Optional[object]) -> None:
        """Initialize JESD mode selector page.

        Args:
            state: Application state object
        """
        self.state = state
        self.part_images = {}

    def write(self) -> None:
        """Render the JESD mode selector page."""
        # Get supported parts that have quick_configuration_modes
        import adijif  # noqa: F401

        supported_parts_filtered = []
        for part in sp:
            try:
                converter = eval(f"adijif.{part}()")  # noqa: S307
                qsm = converter.quick_configuration_modes  # noqa: F841
                supported_parts_filtered.append(part)
            except Exception:  # noqa: S110
                pass
        supported_parts = supported_parts_filtered

        st.title("JESD204 Mode Selector")

        @st.dialog("About JESD204 Mode Selector")
        def help() -> None:
            """Display help dialog."""
            st.write(  # noqa: S608
                "This tool helps you select the appropriate JESD204 mode for "
                "your application. It supports both ADCs, DACs, MxFEs, and "
                "Transceivers from the ADI portfolio. Modeling their JESD204 "
                "mode tables and clocking limitations of the individual "
                "devices.\n\n"
                "To use the tool, select a part from the dropdown menu. You "
                "can then configure the datapath settings such as "
                "decimation/interpolation and the converter sample rate. The "
                "tool will derive the necessary clocks for the selected "
                "configuration. Filter different JESD204 parameters to find a "
                "suitable mode for your application. The valid modes will be "
                "displayed in a table, along with the derived settings."
                "\n\n"
                "JESD204 settings can be exported as CSV from the table."
            )

        st.button("Help", on_click=help)

        sb = st.selectbox(
            label="Select a part",
            options=supported_parts,
            format_func=lambda x: x.upper(),
            key="converter_part_select",
        )

        converter = eval(f"adijif.{sb}()")  # noqa: S307

        # Show diagram
        with st.expander(label="Diagram", expanded=True):
            if sb not in self.part_images:
                if converter.converter_type.lower() == "adc":
                    self.part_images[sb] = draw_adc(converter)
                    # FIXME: State is not being saved
                else:
                    self.part_images[sb] = draw_dac(converter)
            st.image(self.part_images[sb], width="stretch")

        # Datapath Configuration
        with st.expander("Datapath Configuration", expanded=True):
            decimation = gen_datapath(converter)

            scalar = st.selectbox("Units", options=["Hz", "kHz", "MHz", "GHz"], index=3)
            if scalar == "Hz":
                scalar_value = 1
                new_default = 1e9
                min_value = 1
                max_value = 28e9
            elif scalar == "kHz":
                scalar_value = 1e3
                new_default = 1e6
                min_value = 100
                max_value = 28e6
            elif scalar == "MHz":
                scalar_value = 1e6
                new_default = 1e3
                min_value = 1e3
                max_value = 28e3
            elif scalar == "GHz":
                scalar_value = 1e9
                new_default = 1
                min_value = 0.1
                max_value = 28

            min_value = float(min_value)
            max_value = float(max_value)
            new_default = float(new_default)

            converter_rate = st.number_input(
                f"Converter Rate ({scalar})",
                value=new_default,
                min_value=min_value,
                max_value=max_value,
            )
            converter_rate = scalar_value * converter_rate
            converter.sample_clock = converter_rate / decimation

        # Derived settings
        dict_data = {
            "Derived Setting": ["Sample Rate (MSPS)"],
            "Value": [converter.sample_clock / 1e6],
        }
        df = pd.DataFrame.from_dict(dict_data)
        st.dataframe(df, hide_index=True, width="stretch")

        cols = st.columns(2, border=True)

        # JESD204 Configuration Inputs
        options, all_modes = get_jesd_controls(converter)
        selections = {}

        with cols[0]:
            st.subheader("Configuration")

            for option in options:
                selections[option] = st.multiselect(
                    option, options[option]
                )  # noqa: S608

        # Output table of valid modes and calculate clocks
        selections = {k: v for k, v in selections.items() if v != []}
        modes_all_info, found_modes = get_valid_jesd_modes(
            converter, all_modes, selections
        )

        with cols[1]:
            st.subheader("JESD204 Modes")

            if found_modes:
                # Create formatted table of modes

                # Convert to DataFrame so we can change orientation
                df = pd.DataFrame(modes_all_info)

                show_valid = st.toggle("Show only valid modes", value=True)
                if show_valid:
                    df = df[df["Valid"] == "Yes"]
                    df = df.drop(columns=["Valid"])

                # Create new index column and move mode to separate column
                df["Mode"] = df.index
                df = df.reset_index(drop=True)
                # Make mode first column
                cols = df.columns.tolist()
                cols = cols[-1:] + cols[:-1]
                df = df[cols]

                # Change jesd_class column name to be JESD204 Class
                df = df.rename(columns={"jesd_class": "JESD204 Class"})
                df = df.rename(columns={"Mode": "Quickset Mode"})

                # Change data in jesd_class column to be more human readable
                df["JESD204 Class"] = df["JESD204 Class"].replace(
                    {"jesd204b": "204B", "jesd204c": "204C"}
                )

                # Remove any blank rows (rows where all values are NaN or empty)
                df = df.dropna(how="all")
                df = df[df.astype(str).ne("").any(axis=1)]

                to_disable = df.columns
                # Use fixed height with scrolling for consistent display
                # Calculate dynamic height but cap it for better UX
                min_height = 150
                max_height = 600
                row_height = 35
                calculated_height = min(
                    max(len(df) * row_height + 38, min_height), max_height
                )

                st.data_editor(  # noqa: F841
                    df,
                    width="stretch",
                    disabled=to_disable,
                    hide_index=True,
                    height=calculated_height,
                )

            else:
                st.write("No modes found")
